/*
 * MainFrame: メインウィンドウ
 * 注記: ObserverパターンのConcreteObserverクラス 
 *	BridgeパターンのAbstractionクラス
 */

package org.dbc_works.app.sbfiles;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeSet;

import org.dbc_works.lib.environment.EnvInfo;
import org.dbc_works.lib.io.FileNameComparator;
import org.dbc_works.lib.io.JavaFileUtilities;
import org.dbc_works.lib.model.Command;
import org.dbc_works.lib.model.CommandList;
import org.dbc_works.lib.ArgumentInfo;
import org.dbc_works.lib.ui.ActionManager;
import org.dbc_works.lib.ui.MainFrameBase;
import org.dbc_works.lib.util.zip.ZipUtilities;

import org.dbc_works.app.sbfiles.cmd.GZipCommand;
import org.dbc_works.app.sbfiles.cmd.GUnzipCommand;
import org.dbc_works.app.sbfiles.cmd.SetNumberCommand;
import org.dbc_works.app.sbfiles.cmd.SfCommand;
import org.dbc_works.app.sbfiles.cmd.UpdateEntryOrderCommand;
import org.dbc_works.app.sbfiles.view.LabelManager;
import org.dbc_works.app.sbfiles.view.ImplementOperation;
import org.dbc_works.app.sbfiles.view.UpdateHint;

public final class MainFrame extends MainFrameBase
	implements Observer
{
	//
	// 定数
	//
	public static final String PROPKEY_LASTPATH	= "LastPath";

	//
	// 内部型定義
	//
	
	// Observer更新パラメータ 
	public interface UpdateArgument
	{
		int getTotalCount();
		int getProcessCount();
	}
	
	// アクションイベント処理抽象基本クラス
	abstract protected class Action implements ActionListener
	{
		//
		// 実装
		//
		
		// ラベルプロパティを返す
		final protected Properties  getLabelProperties()
		{
			return LabelManager.getLabelProperties();
		}

		//
		// 再定義メソッド(公開)
		//
		
		// アクションを実行する
		final public void actionPerformed(
			ActionEvent event)	// アクションイベント
		{
			if (inProcess != false) {
				beep();
				return;
			}

			try {
				init();
				if (canExecute()) {
					setWaitCursor();
					executeCommand(getCommand());
				}
			}
			catch (Exception e) {
				outputExceptionInfo(e);
				showMessage(e.toString());
				setStatusMessage("");
			}
			finally {
				restoreCursor();
			}
		}
		
		//
		// 再定義可能メソッド(実装)
		//
		
		protected void init()
		{
		}
		
		//
		// 再定義必須メソッド(実装)
		//
		
		abstract protected boolean canExecute();
		abstract protected SfCommand getCommand();
	}

	// 並び順更新処理抽象基本クラス
	abstract protected class UpdateOrderAction extends Action
	{
		//
		// 属性
		//
		
		protected File[] updateEntries;
		protected String[] updateEntryNames;
		
		//
		// 再定義メソッド(実装)
		//
		
		// 実行可能かを返す
		final protected boolean canExecute()
		{
			Properties messages = getLabelProperties(); 			
			if (currentDirectory.exists() == false) {
				showMessage(messages.getProperty(LabelManager.MSG_DIRECTORYISNOTEXIST));
				setPathEntries(getExistDirectory(currentDirectory));
				return false;
			}
			else if (currentDirectory.getParentFile() == null) {
				showMessage(messages.getProperty("Msg_Info_Root"));
				return false;
			}

			String[] currentEntryNames = currentDirectory.list();
			if (Arrays.equals(updateEntryNames, currentEntryNames)) {
				// 変更不要
				return false;
			}
			else if (updateEntryNames.length != currentEntryNames.length) {
				showMessage(messages.getProperty(LabelManager.MSG_ENTRYUPDATED));
				return false;
			}

			Arrays.sort(currentEntryNames);
			for (int nIndex = 0; nIndex < updateEntryNames.length; nIndex++) {
				if (Arrays.binarySearch(currentEntryNames, updateEntryNames[nIndex]) < 0) {
					showMessage(messages.getProperty(LabelManager.MSG_ENTRYUPDATED));
					return false;
				}
			}
			
			return true;
		}

		// 実行コマンドを返す
		final protected SfCommand getCommand()
		{
			return new UpdateEntryOrderCommand(getCommandName(),
												currentDirectory,
												updateEntries,
												MainFrame.this);
		}
		
		//
		// 再定義必須メソッド(実装)
		//
		
		abstract protected String getCommandName();
	}

	//
	// クラス属性(実装)
	//
	private static boolean DEBUG = true;

	private static final MainFrame instance = new MainFrame();

	//
	// クラスメソッド(公開)
	//

	// 唯一のインスタンスを返す
	static MainFrame getInstance()
	{
		return instance;
	}

	//
	// 属性
	//
	
	// ソートアクション
	final ActionListener sortAction	= new UpdateOrderAction() {
		//
		// 再定義メソッド(実装)
		//

		// 初期化する		
		protected void init()
		{
			// ソートしたエントリの並び順を保存する
			final Set sortedDirectoris = new TreeSet(FileNameComparator.instance);
			final Set sortedFiles = new TreeSet(FileNameComparator.instance);
			final File[] originalEntries = currentDirectory.listFiles();
		
			// ディレクトリとファイルに分離しながらソートした集合を作成する
			for (int nIndex= 0; nIndex < originalEntries.length; nIndex++) {
				if (originalEntries[nIndex].isDirectory()) {
					sortedDirectoris.add(originalEntries[nIndex]);
				}
				else {
					sortedFiles.add(originalEntries[nIndex]);
				}
			}
			List list = new ArrayList();
			list.addAll(sortedDirectoris);
			list.addAll(sortedFiles);

			updateEntries = (File[])list.toArray(new File[list.size()]);

			updateEntryNames = new String[updateEntries.length];
			for (int nIndex = 0; nIndex < updateEntries.length; nIndex++) {
				updateEntryNames[nIndex] = updateEntries[nIndex].getName();
			}
		}
		
		// コマンド名を返す
		protected String getCommandName()
		{
			return LabelManager.getCommandNameProperties().getProperty(LabelManager.COMMAND_SORT);
		}
	};

	// 並び順更新アクション
	final ActionListener updateOrderAction	= new UpdateOrderAction() {
		//
		// 再定義メソッド(実装)
		//
		
		// 初期化する		
		protected void init()
		{
			updateEntryNames = getOperation().getEntryNames();

			updateEntries = new File[updateEntryNames.length];
			for (int nIndex = 0; nIndex < updateEntries.length; nIndex++) {
				updateEntries[nIndex] = new File(currentDirectory, updateEntryNames[nIndex]);
			}
		}
		
		// コマンド名を返す
		protected String getCommandName()
		{
			return LabelManager.getCommandNameProperties().getProperty(LabelManager.COMMAND_UPDATE_ORDER);
		}
	};

	// 連番設定アクション
	final ActionListener numberingAction	= new ActionListener() {
		// アクションを実行する
		final public void actionPerformed(
			ActionEvent event)	// アクションイベント
		{
			EventQueue.invokeLater(new Runnable() {
				public void run() {
					setNumbering();
				}
			});
		}
	};
	
	// GZIP圧縮アクション
	final ActionListener gzipAction	= new Action() {
		//
		// 属性
		//
		
		private File[] gzipEntries;
		
		//
		// 再定義メソッド(実装)
		//
		
		// 初期化する		
		protected void init()
		{
			String[] gzipEntryNames = getOperation().getSelectedEntryNames();
			gzipEntries = new File[gzipEntryNames.length];
			for (int nIndex = 0; nIndex < gzipEntryNames.length; nIndex++) {
				gzipEntries[nIndex] = new File(currentDirectory, gzipEntryNames[nIndex]);
			}
		}
		
		// 実行可能かを返す
		final protected boolean canExecute()
		{
			Properties messages = getLabelProperties(); 			
			if (currentDirectory.exists() == false) {
				showMessage(messages.getProperty(LabelManager.MSG_DIRECTORYISNOTEXIST));
				setPathEntries(getExistDirectory(currentDirectory));
				return false;
			}
			if (gzipEntries.length == 0) {
				showMessage(messages.getProperty(LabelManager.MSG_NOTSELECTED));
				return false;
			}
			
			for (int nIndex = 0; nIndex < gzipEntries.length; nIndex++) {
				File file = gzipEntries[nIndex];
				if (file.exists() == false) {
					throw new IllegalStateException();
				}
				if (file.isDirectory()) {
					// ディレクトリは処理対象外
					showMessage(messages.getProperty("Msg_Info_DirectoryCannotGZip"));
					return false;
				}
				else if (ZipUtilities.getFileType(file) == ZipUtilities.FileType.GZIP) {
					// GZIPファイルの再圧縮は処理対象外
					showMessage(messages.getProperty("Msg_Info_IncludeGZipedFile"));
					return false;
				}
			}

			return true;
		}

		// 実行コマンドを返す
		protected SfCommand getCommand()
		{
			return new GZipCommand(currentDirectory, gzipEntries, MainFrame.this);
		}
	};
	
	// GZIP伸長アクション
	final ActionListener gunzipAction	= new Action() {
		//
		// 属性
		//
		
		private File[] gzipEntries;
		
		//
		// 再定義メソッド(実装)
		//
		
		// 初期化する		
		protected void init()
		{
			String[] gzipEntryNames = getOperation().getSelectedEntryNames();
			gzipEntries = new File[gzipEntryNames.length];
			for (int nIndex = 0; nIndex < gzipEntryNames.length; nIndex++) {
				gzipEntries[nIndex] = new File(currentDirectory, gzipEntryNames[nIndex]);
			}
		}
		
		// 実行可能かを返す
		final protected boolean canExecute()
		{
			Properties messages = getLabelProperties(); 			
			if (currentDirectory.exists() == false) {
				showMessage(messages.getProperty(LabelManager.MSG_DIRECTORYISNOTEXIST));
				setPathEntries(getExistDirectory(currentDirectory));
				return false;
			}
			if (gzipEntries.length == 0) {
				showMessage(messages.getProperty(LabelManager.MSG_NOTSELECTED));
				return false;
			}
			
			for (int nIndex = 0; nIndex < gzipEntries.length; nIndex++) {
				File file = gzipEntries[nIndex];
				if (file.exists() == false) {
					throw new IllegalStateException();
				}
				if (file.isDirectory()
				|| ZipUtilities.getFileType(file) != ZipUtilities.FileType.GZIP) {
					showMessage(messages.getProperty("Msg_Info_IncludeNonGZipedFile"));
					return false;
				}
			}

			return true;
		}

		// 実行コマンドを返す
		protected SfCommand getCommand()
		{
			return new GUnzipCommand(currentDirectory, gzipEntries, MainFrame.this);
		}
	};
	
	// 取り消しアクション
	final ActionListener undoAction	= new ActionListener() {
		// アクションを実行する
		final public void actionPerformed(
			ActionEvent event)	// アクションイベント
		{
			if (inProcess != false) {
				beep();
				return;
			}

			inProcess = true;
			try {
				if (commands.canUndo()) {
					setWaitCursor();
					Command.Result result = commands.undo();
					reportResult(result);
	
					if (result == SfCommand.Result.SUCCESS) {
						setStatusMessage(LabelManager.getLabelProperties().getProperty("Status_Undo"));
					}
				}
				else {
					showMessage(LabelManager.getLabelProperties().getProperty("Msg_Info_CannotUndo"));
				}
			}
			catch (Exception e) {
				outputExceptionInfo(e);
				showMessage(e.toString());
				setStatusMessage("");
				commands.setFixed();
			}
			finally {
				setActionStatus();
				restoreCursor();
				inProcess = false;
			}
		}
	};

	// やりなおしアクション
	final ActionListener redoAction	= new ActionListener() {
		// アクションを実行する
		final public void actionPerformed(
			ActionEvent event)	// アクションイベント
		{
			if (inProcess != false) {
				beep();
				return;
			}

			inProcess = true;
			try {
				if (commands.canRedo()) {
					setWaitCursor();
					Command.Result result = commands.redo();
					reportResult(result);
	
					if (result == SfCommand.Result.SUCCESS) {
						setStatusMessage(LabelManager.getLabelProperties().getProperty("Status_Redo"));
					}
				}
				else {
					showMessage(LabelManager.getLabelProperties().getProperty("Msg_Info_CannotRedo"));
				}
			}
			catch (Exception e) {
				outputExceptionInfo(e);
				showMessage(e.toString());
				setStatusMessage("");
				commands.removeRedoCommand();
			}
			finally {
				setActionStatus();
				restoreCursor();
				inProcess = false;
			}
		}
	};
 
	private CommandList commands = new CommandList();
	
	private Frame	frameImp;

	private File	currentDirectory;
	private File	userDirectory;
	private File	userRootDirectory;

	private ActionManager actionManager;

	private boolean	inProcess;
	private Timer	statusMessageResetTimer;

	//
	// コンストラクタ
	//
	private MainFrame()
	{
		super(true);
	}

	//
	// 操作
	//

	// ビュー更新を処理する
	public void updateView(
		UpdateHint hint)	// 更新ヒント
	{
		File fileSet = null;

		setWaitCursor();
		try {
			if (hint.isRoot) {
				// ルート更新
				if (hint.name != null) {
					// 選択
					fileSet = new File(hint.name);
				}
				else {
					// ルートリスト更新
					getOperation().setRoots(File.listRoots());
					setStatusMessage(LabelManager.getLabelProperties().getProperty("Status_UpdateRoot"));
					fileSet = currentDirectory;
				}
			}
			else {
				// 子ディレクトリ選択
				fileSet = new File(currentDirectory, hint.name);
			}

			if (fileSet.exists() == false) {
				Properties messages = LabelManager.getLabelProperties(); 			
				showMessage(messages.getProperty(LabelManager.MSG_DIRECTORYISNOTEXIST));
				fileSet = getExistDirectory(currentDirectory);
			}
		}
		catch (Exception e) {
			outputExceptionInfo(e);
			showMessage(e.toString());
			fileSet = currentDirectory;
		}
		finally {
			setPathEntries(fileSet);
			restoreCursor();
		}
	}

	// エントリを設定する
	public boolean setPathEntries(
		File filePath)	// 設定パス
	{
		if (filePath == null) {
			throw new NullPointerException();
		}
		
		boolean bSuccess = true;
		try {
			getOperation().setPathEntries(filePath, filePath.listFiles());
			currentDirectory = filePath;
		}
		catch (Exception e) {
			outputExceptionInfo(e);
			showMessage(e.toString());
			if (filePath != currentDirectory) {
				setPathEntries(currentDirectory);
			}
			bSuccess = false;
		}
		return bSuccess;
	}

	//
	// 実装
	//

	// 実装を初期化する
	private void initImplementor(
		Frame frame)	// 実装(フレーム)
	{
		final EnvInfo info = EnvInfo.getInstance();
		frame.setTitle(getAppName(false));

		// フォントを設定する
		Font fontBase = info.getBaseFont(frame);
		final String strLang = System.getProperties().getProperty("user.language");
		if (File.separatorChar == '\\'
		&& String.CASE_INSENSITIVE_ORDER.compare(strLang, "ja") == 0) {
			// Windowsでバックスラッシュを'\'として表示するには"Monospaced"でないと
			// 駄目な模様
			fontBase = new Font("Monospaced", fontBase.getStyle(), fontBase.getSize()); 
		}
		frame.setFont(fontBase);

		// メニューとツールバーを作成する
		final ControllerDefinition definition = new ControllerDefinition();
		final Component[] tools = definition.buildMainController(this, frame, actionManager);

		final ImplementOperation op = (ImplementOperation)frame;
		op.setAbstraction(this);
		op.initComponent();
		op.addToolBarComponents(tools);
		op.setStatusMessage("");
		op.setMoveParentActionListener(new ActionListener() {
			// アクションを実行する
			public void actionPerformed(
				ActionEvent e)	// アクションイベント
			{
				final File fileParent = currentDirectory.getParentFile(); 
				if (fileParent != null) {
					final File fileSetPath = getExistDirectory(fileParent);
					if (fileSetPath != fileParent) { 
						Properties messages = LabelManager.getLabelProperties(); 			
						showMessage(messages.getProperty(LabelManager.MSG_DIRECTORYISNOTEXIST));
					}

					setWaitCursor();
					setPathEntries(fileSetPath);
					restoreCursor();
				}
				else {
					beep();
				}
			}
		});
		op.setUpdateOrderActionListener(updateOrderAction);

		setActionStatus();
	}

	// フレームをImplementOperationインターフェースとして返す
	private ImplementOperation getOperation()
	{
		return (ImplementOperation)getImplementor();
	}

	// 存在するディレクトリを返す
	private File getExistDirectory(
		File fileSetPath)	// 設定予定のパス
	{
		File fileExists = fileSetPath;
		if (fileExists == null || fileExists.exists() == false) {
			fileExists = (userDirectory.exists() ? userDirectory : userRootDirectory);
		}
		return fileExists;
	}

	// 現在のディレクトリが変更されたかを返す
	private boolean isCurrentDirectoryChanged(
		String[] entries)	// 保持しているエントリ名
	{
		final String[] sortedEntries = new String[entries.length];
		System.arraycopy(entries, 0, sortedEntries, 0, entries.length);
		Arrays.sort(sortedEntries);
		final String[] sortedDirList = currentDirectory.list();
		Arrays.sort(sortedDirList);
		return (Arrays.equals(sortedEntries, sortedDirList) == false);
	}
	
	// 状態メッセージを設定する
	private void setStatusMessage(
		String strMessage)	// 設定メッセージ
	{
		if (0 < strMessage.length()) {
			beep();
			if (statusMessageResetTimer != null) {
				statusMessageResetTimer.cancel();
			}
			
			statusMessageResetTimer = new Timer();
			statusMessageResetTimer.schedule(
				new TimerTask() {
					public void run() {
						setStatusMessage("");
					}
				},
				3 * 1000);
		}
		getOperation().setStatusMessage(strMessage);
	}

	// 結果を報告する
	private void reportResult(
		Command.Result result)	// 結果
	{
		Properties messages = LabelManager.getLabelProperties();
		 			
		if (result == SfCommand.Result.FAIL_CREATE_TEMP_DIR) {
			showMessage(messages.getProperty("Msg_Info_FailCreateTempDir"));
		}
		else if (result == SfCommand.Result.FAIL_MOVE_ENTRY) {
			showMessage(messages.getProperty("Msg_Info_FailMoveEntry"));
		}
		else if (result == SfCommand.Result.FAIL_DELETE_TEMP_DIR) {
			showMessage(messages.getProperty("Msg_Info_FailDeleteTempDir"));
		}
		else if (result != SfCommand.Result.SUCCESS) {
			outputToError(result.name);
			showMessage(messages.getProperty("Msg_Info_UnknownError")
						+ result.name);
		}
	}

	// アクションの状態を設定する
	private void setActionStatus()
	{
		actionManager.setEnable(undoAction, commands.canUndo());
		actionManager.setEnable(redoAction, commands.canRedo());
	}
	
	// コマンドを実行する
	private boolean executeCommand(
		SfCommand command)	// 実行コマンド
		throws Exception
	{
		if (command == null) {
			throw new NullPointerException();
		}
		
		Command.Result result;
		inProcess = true;
		try {	
			result = commands.execute(command);
			reportResult(result);
		
			if (result == SfCommand.Result.SUCCESS) {
				setStatusMessage(LabelManager.getLabelProperties().getProperty("Status_Success"));
			}
			else {
				setStatusMessage("");
			}
		}
		finally {
			setActionStatus();
			inProcess = false;
		}
		
		return (result == SfCommand.Result.SUCCESS);
	}
	
	// 連番設定を実行する
	private void setNumbering()
	{
		if (inProcess != false) {
			beep();
			return;
		}
		
		ImplementOperation op = getOperation();
		NumberingSetting setting;
		try {
			// 設定ダイアログを表示する
			setting = op.showNumberingSettingDialog(getFrameTitle("NumberingSettingDialog"));
			if (setting == null) {
				return;
			}

			// 対象を取得する
			String[] entries;
			int nSelected = op.getSelectionCount();
			if (1 < nSelected) {
				// 複数選択状態
				// 選択を取得する
				entries = op.getSelectedEntryNames();
	
				// ディレクトリとファイルが混在していたら処理しない
				int nFile = 0;
				int nDirectory = 0;
				for (int nIndex = 0; nIndex < entries.length; nIndex++) {
					File file = new File(currentDirectory, entries[nIndex]);
					if (file.exists() == false) {
						throw new IllegalStateException();
					}
					if (file.isDirectory()) {
						nDirectory++;
					}
					else {
						nFile++;
					}
				}
				if (0 < nDirectory && 0 < nFile) {
					// メッセージ表示
					showMessage(LabelManager.getLabelProperties().getProperty("Msg_Info_CannotFileAndDirOperation"));
					return;
				}
			}
			else {
				// 単数選択状態
				// 選択にかかわらずファイルのみを取得する
				entries = op.getEntryNames(false);
			}

			setWaitCursor();
			SetNumberCommand cmd = new SetNumberCommand(currentDirectory, entries, setting, this); 
			if (executeCommand(cmd) != false) {
				setPathEntries(currentDirectory);
			}
		}
		catch (Exception e) {
			outputExceptionInfo(e);
			showMessage(e.toString());
			setStatusMessage("");
		}
		finally {
			restoreCursor();
		}
	}
	
	//
	// 再定義メソッド(実装)
	//

	// アプリケーション名を返す	
	protected String getAppName(
		boolean bShort)	// 短縮名フラグ
	{
		return (bShort != false ? "sbfiles" : "StompBox/Files");
	}

	// 実装を作成して返す
	// 注記: J2MEに存在しないクラスのロードを避けるためにリフレクションで作成する
	protected Frame createImplementor(
		ArgumentInfo argument)	// 引数情報
	{
		Frame frame = null;

		try {
			final boolean bAwt = isToolkitAwt(argument);
			if (bAwt == false) {
				initSwing(argument);
			//	frame = new swing.MainFrameImp();
				frame = (Frame)Class.forName("org.dbc_works.app.sbfiles.view.swing.MainFrameImp").newInstance();
			}
			else {
				frame = new org.dbc_works.app.sbfiles.view.awt.MainFrameImp();
			//	frame = (Frame)Class.forName("org.dbc_works.app.sbfiles.view.awt.MainFrameImp").newInstance();
			}
			actionManager = ActionManager.createInstance(bAwt);

			initImplementor(frame);
		}
		catch (ClassNotFoundException exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}
		catch (IllegalAccessException exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}
		catch (InstantiationException exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}

		return frame;		
	}

	// デフォルトのサイズを返す
	protected Dimension getDefaultSize()
	{
		Frame frame = getImplementor();
		if (frame == null) {
			throw new IllegalStateException();
		}

		Dimension screenSize = frame.getToolkit().getScreenSize();
		return new Dimension(screenSize.width / 3, screenSize.height / 2);
	}

	// 初期化する
	protected boolean doInit(
		ArgumentInfo argument)	// 引数情報
	{
		setTitle(getAppName(false) + " Ver.0.4");
		setIcon();

		final EnvInfo info = EnvInfo.getInstance();
		if (info.isMultiRootFileSystem()) {
			getOperation().setRoots(File.listRoots());
		}

		userDirectory = new File(info.getUserHomeDir());
		userRootDirectory = JavaFileUtilities.getRootFile(userDirectory); 

		final Map properties = getStateProperties();
		final String strLastPath = (String)properties.get(PROPKEY_LASTPATH);
		final File filePath = (strLastPath != null ? new File(strLastPath) : userDirectory); 
		setPathEntries(getExistDirectory(filePath));
		return true;
	}

	// 追加状態を復元する
	protected void restoreAdditionalState()
	{
		try {
			getOperation().setAdditionalState(getStateProperties());
		}
		catch (Exception exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}
	}

	// 追加状態を設定する
	protected void setAdditionalStatePropreties()
	{
		setStatePropertyValue(PROPKEY_LASTPATH, currentDirectory.getAbsolutePath()); 

		Map mapAdditionalState = getOperation().getAdditionalState();
		if (mapAdditionalState != null) {
			Set keySet = mapAdditionalState.keySet();
			Iterator it = keySet.iterator();
			while (it.hasNext()) {
				String strKey = (String)it.next();
				setStatePropertyValue(strKey, (String)mapAdditionalState.get(strKey));
			}
		}
	}

	// 終了可能かを返す
	protected synchronized boolean canExit()
	{
		if (inProcess != false) {
			beep();
		}
		return (inProcess == false);
	}

	// 終了処理を行う
	protected void prepareExit()
	{
		// 特に何もしない
	}

	//
	// 再定義メソッド(WindowAdapter)
	//

	// アクティブ化処理
	public void windowActivated(
		WindowEvent e)	// イベント情報
	{
		if (e.getID() == WindowEvent.WINDOW_ACTIVATED
		&& isCurrentDirectoryChanged(getOperation().getEntryNames())) {
			// エントリのリストを更新する 
			setWaitCursor();
			setPathEntries(getExistDirectory(currentDirectory));
			restoreCursor();
		}
	}

	//
	// 再定義メソッド(Observer)
	//

	// 更新状態を受けとる
	public void update(
		Observable observable,			// 全件数
		Object argument)		// 進行中件数
	{
		UpdateArgument param = (UpdateArgument)argument;
		getOperation().setProgressInfo(param);
	}
}
