/*
 * SetNumberCommand: 連番設定コマンド 
 */

package org.dbc_works.app.sbfiles.cmd;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.dbc_works.lib.io.FileUtilities;
import org.dbc_works.lib.yaml.YamlSerializer;

import org.dbc_works.app.sbfiles.CountReporter;
import org.dbc_works.app.sbfiles.MainFrame;
import org.dbc_works.app.sbfiles.NumberingSetting;
import org.dbc_works.app.sbfiles.view.LabelManager;

public final class SetNumberCommand extends ProcessEachEntryCommand
{
	//
	// 内部型定義
	//

	// 変換結果
	static private class RenameResult
	{
		//
		// 属性
		//
		
		final File originalFile;
		final File renamedFile;
		
		//
		// コンストラクタ
		//
		RenameResult(
			File fileOriginal,	// オリジナルのファイル
			File fileRenamed)	// 変換後のファイル
		{
			if (fileOriginal == null) {
				throw new NullPointerException();
			}
			else if (fileOriginal.exists()) {
				throw new IllegalArgumentException();
			}
			if (fileRenamed == null) {
				throw new NullPointerException();
			}
			else if (fileRenamed.exists() == false) {
				throw new IllegalArgumentException();
			}

			originalFile = fileOriginal;
			renamedFile = fileRenamed;
		}
	}

	//
	// クラスメソッド(実装)
	//

	// 連番設定ファイル名を作成する
	private static String makeNumberedEntryName(
		boolean bSetStart,				// 先頭設定フラグ
		String strLeftPad,				// 桁数文字列
		int nNumber,					// 設定番号
		String strOriginalEntryName)	// 元エントリ名
	{
		String strNumber = strLeftPad + Integer.toString(nNumber);
		strNumber = strNumber.substring(strNumber.length() - strLeftPad.length());

		String[] entryNames = FileUtilities.separateNameAndExt(strOriginalEntryName);
		if (bSetStart != false) {
			int nIndex = 0;
			while (Character.isDigit(entryNames[0].charAt(nIndex))) {
				nIndex++;
				if (entryNames[0].length() <= nIndex) {
					break;
				}
			}
			if (entryNames[0].length() <= nIndex) {
				entryNames[0] = strNumber;
			}
			else {
				entryNames[0] = strNumber + entryNames[0].substring(nIndex);
			}
		}
		else {
			int nIndex = entryNames[0].length() - 1;
			while (Character.isDigit(entryNames[0].charAt(nIndex))) {
				nIndex--;
				if (nIndex < 0) {
					break;
				}
			}
			if (nIndex < 0) {
				entryNames[0] = strNumber;
			}
			else {
				entryNames[0] = entryNames[0].substring(0, nIndex + 1) + strNumber;
			}
		}
		return FileUtilities.combineNameAndExt(entryNames);
	}

	// 退避ファイル名を作成する
	private static File makeEscapedFileName(
		File originalFile)	// オリジナルのファイル
	{
		if (originalFile == null) {
			throw new NullPointerException();
		}

		File escapedFile = null;
		final String[] astrName = FileUtilities.separateNameAndExt(originalFile.getName());

		int nIndex = 1;
		while (escapedFile == null || escapedFile.exists()) {
			String strEscapedName = astrName[0]
									 + ".original";
			if (1 < nIndex) {									 
				strEscapedName += "-" + Integer.toString(nIndex);
			}
			escapedFile = new File(originalFile.getParentFile(),
									FileUtilities.combineNameAndExt(strEscapedName, astrName[1]));
			nIndex++;
		}

		return escapedFile;
	}
	
	// 名前変更結果を報告する
	private static void reportRenameResult(
		File fileFrom,					// 変更前のファイル
		File fileTo,					// 変更後のファイル
		YamlSerializer logSerializer)	// ログ記録
		throws IOException
	{
		if (fileFrom == null) {
			throw new NullPointerException();
		}
		else if (fileFrom.exists()) {
			throw new IllegalArgumentException();
		}
		if (fileTo == null) {
			throw new NullPointerException();
		}
		else if (fileTo.exists() == false) {
			throw new IllegalArgumentException();
		}
		if (logSerializer == null) {
			throw new NullPointerException();
		}
		
		System.err.println(fileFrom.getName()
							+ " -> "
							+ fileTo.getName());
		logSerializer.writeMapping(fileFrom.getName(), fileTo.getName());
	}

	//
	// 属性
	//

	private final File[] originalFiles;
	private final File[] numberedFiles;
	private final List renamedList;

	//
	// コンストラクタ
	//
	public SetNumberCommand(
		File fileTarget,			// 対象ディレクトリ
		String[] entries,			// 更新エントリ
		NumberingSetting setting,	// 連番設定条件 
		MainFrame frame)			//
	{
		super(LabelManager.getCommandNameProperties().getProperty("Command_SetNumber"),
				fileTarget,
				frame);
		
		if (entries == null) {
			throw new NullPointerException();
		}
		else if (entries.length == 0){
			throw new IllegalArgumentException();
		}
		if (setting == null) {
			throw new NullPointerException();
		}

		originalFiles = new File[entries.length];
		for (int nIndex = 0; nIndex < entries.length; nIndex++) {
			originalFiles[nIndex] = new File(targetDirectory, entries[nIndex]);
			if (originalFiles[nIndex].exists() == false
			|| originalFiles[nIndex].isFile() == false) {
				throw new IllegalArgumentException();
			}
		}

		// 桁数から前置文字列を作成する
		final int nDigit = Integer.toString(setting.start + entries.length).length();
		char[] pads = new char[setting.minLength < nDigit ? nDigit : setting.minLength];
		Arrays.fill(pads, '0');
		String strLeftPad = new String(pads);
		
		// 名前を作成する
		numberedFiles = new File[entries.length];
		for (int nIndex = 0; nIndex < entries.length; nIndex++) {
			final String strNumberedEntry = makeNumberedEntryName(setting.setStart,
																	strLeftPad,
																	setting.start + nIndex,
																	entries[nIndex]);
			numberedFiles[nIndex] = new File(targetDirectory, strNumberedEntry);
		}

		renamedList = new ArrayList();
	}

	//
	// 再定義メソッド(Command)
	//

	// 実行可能かを返す
	// 注記: 取り消しの判定はむずかしいので省略する
	public boolean canExecute()
	{
		boolean bCanExecute = true;
		if (isExecuted() == false) {
			for (int nIndex = 0; nIndex < originalFiles.length; nIndex++) {
				if (originalFiles[nIndex].exists() == false
				|| originalFiles[nIndex].isFile() == false) {
					bCanExecute = false;
					break;
				}
			}
		}
		return bCanExecute;
	}

	//
	// 再定義メソッド(ProcessEachEntryCommand)
	//

	// 処理エントリ数を返す
	protected int getProcessEntryCount()
	{
		return originalFiles.length;
	}
	
	// 処理名を返す
	protected String getOperationName()
	{
		return "numbering";
	}

	// 処理を実行する
	protected void process(
		YamlSerializer logSerializer,	//
		CountReporter reporter)			// 
		throws IOException
	{
		for (int nIndex = 0; nIndex < originalFiles.length; nIndex++) {
			if (originalFiles[nIndex].exists() == false
			|| originalFiles[nIndex].isFile() == false) {
				throw new IllegalArgumentException();
			}

			// 変換する必要があるかチェックする
			if (originalFiles[nIndex].compareTo(numberedFiles[nIndex]) != 0) {
				// 変換後のファイルがすでに存在したら名前を変える
				if (numberedFiles[nIndex].exists()) {
					File fileEscaped = makeEscapedFileName(numberedFiles[nIndex]);
					if (numberedFiles[nIndex].renameTo(fileEscaped) == false) {
						throw new IOException();
					}
					reportRenameResult(numberedFiles[nIndex], fileEscaped, logSerializer);
					renamedList.add(new RenameResult(numberedFiles[nIndex],
														fileEscaped));
				}
	
				// 変換するファイルがこれまでの処理で変換されているかチェックする
				File fileRename = originalFiles[nIndex];
				Iterator it = renamedList.iterator();
				while (it.hasNext()) { 
					final RenameResult result = (RenameResult)it.next();
					if (result.originalFile.compareTo(fileRename) == 0) {
						fileRename = result.renamedFile;
						break;						
					}
				}
					
				if (fileRename.renameTo(numberedFiles[nIndex]) == false) {
					throw new IOException();
				}
				reportRenameResult(fileRename, numberedFiles[nIndex], logSerializer);
				renamedList.add(new RenameResult(fileRename, numberedFiles[nIndex]));
			}

			reporter.reportCountUp();
		}
	}
	
	// 処理をキャンセルする
	protected void cancelProcess(
		YamlSerializer logSerializer,	//
		CountReporter reporter)			// 
		throws IOException
	{
		while (renamedList.isEmpty() == false) {
			final RenameResult result = (RenameResult)renamedList.remove(renamedList.size() - 1);
			if (result.renamedFile.renameTo(result.originalFile) != false) {
				reportRenameResult(result.renamedFile, result.originalFile, logSerializer);
			}
			if (reporter != null) {
				reporter.reportCountUp();
			}
		}
	}
}
