/*
 * EnvInfo: 環境情報抽象基本クラス
 */

package org.dbc_works.lib.environment;

import org.dbc_works.lib.io.JavaFileUtilities;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

public abstract class EnvInfo
{
	//
	// クラス属性
	//
	private static EnvInfo instance;

	public static final String NEWLINE	= System.getProperty("line.separator");

	protected static final String FONTFACE_DEFAULT = "SansSerif";
	protected static final String[] ARRAY_EMPTY_STRING	= new String[0];

	//
	// クラス初期化子
	//
	
	static
	{
		try {
			// OS名に"Windows"が含まれていればWindows、
			// Zaurusのユーザーディレクトリが存在してjava.homeがJ2MEならZaurus
			// それ以外はUnix系一般と判断する
			// (これだとMacが区別できないな……)
			final String strZaurusUserPath = File.separator
											 + "home"
											 + File.separator
											 + "zaurus";
			if (0 <= System.getProperty("os.name").indexOf("Windows")) {
				instance = new WindowsInfo();
			}
		//	else if (0 <= System.getProperty("user.home").indexOf("zaurus")) {
			else if (new File(strZaurusUserPath).exists()
			&& String.CASE_INSENSITIVE_ORDER.compare(System.getProperty("java.home"), "/home/QtPalmtop/j2me") == 0) {
				// ユーザー名が"zaurus"ならJ2ME搭載Linux Zaurus、
				// そうでなければJeode搭載Linux Zaurusと切り分けできるようだ
			//	if (System.getProperty("user.name").equalsIgnoreCase("zaurus")) {
			//		instance = new ZaurusJ2MEInfo();
			//	}
			//	else {
			//		instance = new ZaurusJ2MEInfo();
			//	}
				instance = new ZaurusJ2MEInfo();
			}
			else {
				instance = new GenericUnixInfo();
			}
		//	System.err.println("Environment: " + instance.getClass().getName());
		}
		catch (Exception exp) {
			instance = new WindowsInfo();
		}
	}

	//
	// クラスメソッド(公開)
	//

	// 適切なインスタンスを返す
	public static final EnvInfo getInstance()
	{
		return instance;
	}

	// クラスパス文字列を作成して返す
	public static final String makeJavaClassPathLine(
		Collection colClassPath)	// クラスパスのリスト
	{
		if (colClassPath == null) {
			throw new NullPointerException();
		}

		StringBuffer sb = new StringBuffer();
		Iterator it = colClassPath.iterator();
		while (it.hasNext()) {
			if (0 < sb.length()) {
				sb.append(File.pathSeparator);
			}
			sb.append(it.next());
		}

		return sb.toString();
	}
	public static final String makeJavaClassPathLine(
		String strStartClassPath,	// 実行クラスのパス
		Collection colClassPath)	// クラスパスのリスト
	{
		if (strStartClassPath == null) {
			throw new NullPointerException();
		}

		Set setTotalClassPath = new TreeSet();
		if (colClassPath != null && colClassPath.isEmpty() == false) {
			setTotalClassPath.addAll(colClassPath);
		}
		if (0 < strStartClassPath.length()
		&& colClassPath.contains(strStartClassPath) == false) {
			setTotalClassPath.add(strStartClassPath);
		}

		return makeJavaClassPathLine(setTotalClassPath);
	}

	//
	// クラスメソッド(実装)
	//

	// ファイルパスからコマンドライン情報を作成して返す
	protected static final String[] getCmdLineInfoFromFile(
		String strFilePath)	// ファイルパス
	{
		if (strFilePath == null) {
			throw new NullPointerException();
		}
		else if (strFilePath.length() == 0) {
			throw new IllegalArgumentException();
		}

		String[] astrArgs = new String[2];
		
		File file = new File(strFilePath).getAbsoluteFile();
		astrArgs[0] = file.getName();
		int nExtIndex = astrArgs[0].lastIndexOf('.');
		if (0 < nExtIndex) {
			astrArgs[0] = astrArgs[0].substring(0, nExtIndex);
		}

		astrArgs[1] = file.getParent();

		return astrArgs;
	}

	//
	// コンストラクタ
	//
	protected EnvInfo()
	{
	}

	//
	// 状態
	//

	// ユーザーのホームディレクトリを返す
	public final String getUserHomeDir()
	{
		return System.getProperty("user.home");
	}

	// アルファチャンネルが有効かを返す
	public final boolean isAlphaCompositeRestricted()
	{
		return Boolean.getBoolean("java.awt.AlphaComposite.SRC_OVER.isRestricted");
	}

	//
	// 操作
	//

	// Javaファイル実行のためのコマンドラインを作成する
	// 標準的なコマンドライン: java [-jar ][-options] class [args]
	// Zaurus J2MEのコマンドライン: /home/QtPalmtop/j2me/bin/cvm -XappName=(name) -cp file:(classpath) (main class name)
	public final String makeJavaCmdLine(
		String strRunJavaFile,		// 実行ファイル
		Collection colClassPath,	// クラスパス
		String strArgument,			// 引数		
		String strName)				// 名前
		throws IOException
	{
		if (strRunJavaFile == null) {
			throw new NullPointerException();
		}
		else if (strRunJavaFile.length() == 0) {
			throw new IllegalArgumentException();
		}
	
		// Javaコマンドのパスを作成する
		// 標準的なコマンドライン: ($java.home)/bin/java
		// Zaurus J2MEのコマンドライン: ($java.home)/bin/cvm -XappName=(name) -cp file:(classpath) (main class name)
		StringBuffer sb = new StringBuffer(System.getProperty("java.home"));
		sb.append(File.separator);
		sb.append("bin");
		sb.append(File.separator);
		sb.append(getJavaCmdName());

		// Javaコマンドの引数を追加する
		sb.append(' ');
		String strArg = makeJavaCmdArgument(strRunJavaFile,
											colClassPath,
											strName);
		sb.append(strArg);

		// Javaプログラム自体への引数を追加する
		if (strArgument != null && 0 < strArgument.length()) {
			sb.append(' ');
			sb.append(strArgument);
		}

		return sb.toString();
	}

	// スクリプトファイルを作成する
	public final void createCommandScriptFile(
		String strCommandLine,	// 作成コマンドライン
		String strFilePath)		// Java実行プログラムフラグ
		throws IOException
	{
		if (strCommandLine == null) {
			throw new NullPointerException();
		}
		else if (strCommandLine.length() == 0) {
			throw new IllegalArgumentException();
		}
		if (strFilePath == null) {
			throw new NullPointerException();
		}
		else if (strFilePath.length() == 0) {
			throw new IllegalArgumentException();
		}
		
		FileOutputStream fos = null;
		OutputStreamWriter sw = null;
		BufferedWriter writer = null;
		try {
			fos = new FileOutputStream(strFilePath);
			sw = new OutputStreamWriter(fos, getDefaultEncoding());
			writer = new BufferedWriter(sw);

			// コマンドラインを書きこむ
			writer.write(strCommandLine);
			writer.newLine();
		}
		finally {
			if (writer != null) {
				try {
					writer.flush();
					writer.close();
				}
				catch (IOException e) {
					// Do nothing.
				}
			}
			else if (sw != null) {
				try {
					sw.flush();
					sw.close();
				}
				catch (IOException e) {
					// Do nothing.
				}
			}
			else if (fos != null) {
				try {
					fos.flush();
					fos.close();
				}
				catch (IOException e) {
					// Do nothing.
				}
			}
		}
	}
	
	//
	// 再定義可能メソッド(状態)
	//

	// デフォルトの文字符号化方式を返す
	public String getDefaultEncoding()
	{
		return System.getProperty("file.encoding");
	}

	//
	// 再定義必須メソッド(状態)
	//
	public abstract boolean isMultiRootFileSystem();
	public abstract boolean isPathCharCaseInsensitive();
	public abstract String getDefaultScriptExt();

	public abstract boolean isSwingAvailable();
	public abstract boolean isWindowTitleBasicLatinOnly();
	public abstract boolean isIconRequired();
	public abstract boolean isFrameResizable();
	public abstract boolean showMainFrameAsFullScreen();
	public abstract Dimension getDefaultMainFrameSize(Component component);
	public abstract Font getBaseFont(Component component);
	public abstract Font getMenuFont(Component component);

	public abstract boolean hasNoMouse();

	//
	// 再定義可能メソッド(実装)
	//

	// Javaコマンドの名前を返す
	protected String getJavaCmdName()
	{
		return "java";
	}

	// Javaコマンドの引数を作成する
	// 標準的な引数: [-jar ][-options] class
	protected String makeJavaCmdArgument(
		String strRunJavaFile,		// 実行ファイル
		Collection colClassPath,	// クラスパス
		String strName)				// 名前
		throws IOException
	{
		File file = new File(strRunJavaFile);
		if (file.exists() == false) {
			throw new IllegalArgumentException();
		}

		JavaFileUtilities.FileType type = JavaFileUtilities.getFileType(file);
		if (type != JavaFileUtilities.FileType.CLASS
		&& type != JavaFileUtilities.FileType.JAR) {
			throw new IllegalArgumentException();
		}

		StringBuffer sb = new StringBuffer();

		String[] astrCmdLineInfo = getCmdLineInfoFromFile(strRunJavaFile);

		if (type == JavaFileUtilities.FileType.JAR) {
			sb.append("-jar ");
		}

		// クラスパス引数を追加する
		String strClassPathLine = makeJavaClassPathLine(astrCmdLineInfo[1], colClassPath); 
		if (0 < strClassPathLine.length()) {
			sb.append("-cp ");
			sb.append(strClassPathLine);
			sb.append(' ');
		}

		// 名前が指定できれば指定する
		if (strName != null && 0 < strName.length()) {
		//	sbCmdLine.append(" -XappName=" + strName + " ");
		}

		// Javaプログラムを追加する
		sb.append(type == JavaFileUtilities.FileType.JAR
				? strRunJavaFile
				: astrCmdLineInfo[0]);

		return sb.toString();
	}
}
