/*
 * CommandList: 
 * 注記: CommandパターンのInvokerクラス 
 */
package org.dbc_works.lib.model;

import java.util.ArrayList;

final public class CommandList
{
	//
	// 属性
	//

	// コマンドのリスト
	private final ArrayList commandList;

	// 実行済みコマンドのインデックス
	private int executedIndex;

	// 確定インデックス
	private int fixedIndex;

	//
	// コンストラクタ
	//
	public CommandList()
	{
		commandList = new ArrayList();
		executedIndex = fixedIndex = -1;
	}
	
	//
	// 状態
	//

	// 更新されているかを返す
	public boolean isModified()
	{
		return (executedIndex != fixedIndex);
	}
	
	// 取り消し可能かを返す
	public boolean canUndo()
	{
		return (commandList.isEmpty() == false && 0 <= executedIndex)
			&& ((Command)commandList.get(executedIndex)).canExecute();
	}
	
	// 取り消し可能なコマンドの名称を返す
	public String getUndoCommandName()
	{
		if (canUndo() == false) {
			throw new IllegalStateException();
		}

		Command undoCommand = (Command)commandList.get(executedIndex);
		return undoCommand.name;
	}
	
	// 再実行可能かを返す
	public boolean canRedo()
	{
		final int nReexecuteIndex = executedIndex + 1;
		return (nReexecuteIndex < commandList.size())
		&& ((Command)commandList.get(nReexecuteIndex)).canExecute();
	}
	
	// 再実行可能なコマンドの名称を返す
	public String getRedoCommandName()
	{
		if (canRedo() == false) {
			throw new IllegalStateException();
		}
		
		return ((Command)commandList.get(executedIndex + 1)).name;
	}

	//
	// 操作
	//

	// 与えられたコマンドを実行する
	public Command.Result execute(
		Command execCommand)	// 実行コマンド
		throws Exception
	{
		if (execCommand == null) {
			throw new NullPointerException();
		}
		else if (execCommand.canExecute() == false) {
			throw new IllegalArgumentException();
		}

		Command.Result result = execCommand.execute(); 
		if (result == Command.Result.SUCCESS) {
			removeRedoCommand();
	
			// 無修正と指定されたコマンドが削除されたら常に修正状態と考える
			if (executedIndex < fixedIndex) {
				fixedIndex = -1;
			}
	
			while (true) {
				try {
					// 実行したコマンドを追加する
					commandList.add(execCommand);
					break;
				}
				catch (OutOfMemoryError error) {
					// 古いコマンドを削除してやりなおす
					// (動作未確認)
					if (commandList.isEmpty()) {
						throw error;
					}
	
					commandList.remove(0);
					commandList.trimToSize();
					executedIndex--;
					if (0 <= fixedIndex) {
						fixedIndex--;
					}
				}
			}
			executedIndex++;
	
			if (executedIndex != (commandList.size() - 1)) {
				throw new IllegalStateException();
			}
		}

		return result;
	}
	
	// 直前に与えられたコマンドの実行を取り消す
	public Command.Result undo()
		throws Exception
	{
		if (canUndo() == false) {
			throw new IllegalStateException("Has no undo command.");
		}

		Command undoCommand = (Command)commandList.get(executedIndex);
		final Command.Result result = undoCommand.cancel(); 
		executedIndex--;
		return result;
	}
	
	// 取り消した直前のコマンドを再実行する
	// 注記: 実装はもっと考えられることがある
	public Command.Result redo()
	throws Exception
	{
		if (canRedo() == false) {
			throw new IllegalStateException();
		}

		Command redoCommand = (Command)commandList.get(executedIndex + 1);
		final Command.Result result = redoCommand.execute();
		executedIndex++;
		return result;
	}

	// コマンドリストをクリアする
	public void clear()
	{
		commandList.clear();
		executedIndex = fixedIndex = -1;
	}
	
	// 取り消したコマンドを削除する
	// 注記: やり直しがうまくいかなかったとき用
	public void removeRedoCommand()
	{
		while ((executedIndex + 1) < commandList.size()) {
			commandList.remove(executedIndex + 1);
		}
		commandList.trimToSize();
	}
	
	// 無修正状態を設定する
	public void setFixed()
	{
		fixedIndex = executedIndex;
	}
}
