/*
 * MainFrameBase: メインフレーム抽象基本クラス
 * (BridgeパターンのAbstractionクラス) 
 * (Template MethodパターンのAbstractClassクラス) 
 */

package org.dbc_works.lib.ui;

import org.dbc_works.lib.ArgumentInfo;
import org.dbc_works.lib.io.FileUtilities;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Point;
import java.awt.datatransfer.Clipboard;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.ResourceBundle;

import org.dbc_works.lib.util.Utilities;
import org.dbc_works.lib.environment.EnvInfo;

abstract public class MainFrameBase extends WindowAdapter
{
	//
	// 定数
	//

	private static final String PROPKEY_POS_LEFT	= "MainFramePosX";
	private static final String PROPKEY_POS_TOP	= "MainFramePosY";
	private static final String PROPKEY_POS_WIDTH	= "MainFrameWidth";
	private static final String PROPKEY_POS_HEIGHT	= "MainFrameHeight";

	//
	// 内部定義型
	//

	// 終了アクション
	public final class CloseAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
		//	implementor.dispatchEvent(new WindowEvent(implementor, WindowEvent.WINDOW_CLOSING));
			windowClosing(null);
		}
	}

	//
	// クラスメソッド(公開)
	//

	// ウィンドウタイトルを返す
	public static final String getFrameTitle(
		String strWindowName)	// ウィンドウ名(入力)
	{
		Locale localeTitle = (EnvInfo.getInstance().isWindowTitleBasicLatinOnly()
							? Locale.ENGLISH
							: Locale.getDefault());
		ResourceBundle resource = ResourceBundle.getBundle("FrameTitleResource",
															localeTitle);
		return resource.getString(strWindowName);
	}

	//
	// クラスメソッド(実装)
	//

	// AWT利用かを返す
	protected static boolean isToolkitAwt(
		ArgumentInfo argument)	// 引数情報
	{
		return (argument.hasOption("awt")
				|| EnvInfo.getInstance().isSwingAvailable() == false);
	}

	// Swingの初期設定を行う
	// 注記: クラスロードを避けるためリフレクションで実行する
	public static void initSwing(
		ArgumentInfo argument)	// 引数情報
	{
		if (argument.hasOption("SwingFrame")) {
			// ウィンドウ修飾を設定する
			// 設定できなくても動作に支障はないので例外はすべて無視する
			String[] astrDecoration = new String[] {
				"javax.swing.JFrame",
				"javax.swing.JDialog",
			};
			try {
			for (int nIndex = 0; nIndex < astrDecoration.length; nIndex++) {
					Class jframe = Class.forName(astrDecoration[nIndex]);
	
					Class[] argTypes = new Class[1];
					argTypes[0] = boolean.class;
					java.lang.reflect.Method method = jframe.getMethod("setDefaultLookAndFeelDecorated",
																		argTypes);
					if (method != null) {
						Object[] args = new Object[1];
						args[0] = Boolean.TRUE;
						method.invoke(null, args);
					}
				}
			}
			catch (ClassNotFoundException exp) {
				// 発生しないはずの例外
				exp.printStackTrace();
			}
			catch (NoSuchMethodException exp) {
				// 1.4より前のJREだと発生する
				exp.printStackTrace();
			}
			catch (IllegalAccessException exp) {
				// 発生しないはずの例外
				exp.printStackTrace();
			}
			catch (java.lang.reflect.InvocationTargetException exp) {
				// 発生しないはずの例外
				exp.printStackTrace();
			}
		}
	}

	//
	// 属性
	//

	public final ActionListener closeAction	= new CloseAction();

	private final Properties properties;

	private Frame implementor;
	private Cursor saveCursor;

	//
	// コンストラクタ
	//
	protected MainFrameBase(
		boolean bUseProperties)	// プロパティファイル使用フラグ
	{
		super();
		properties = (bUseProperties ? new Properties() : null);
	}

	//
	// 状態
	//

	// ログ書きこみストリームライターを返す
	public Writer getLogWriter(
		boolean bByMonth)	// 月単位フラグ
		throws FileNotFoundException
	{
		// ファイル名を作成する
		StringBuffer sb = new StringBuffer();
		sb.append(EnvInfo.getInstance().getUserHomeDir());
		sb.append(File.separatorChar);
		sb.append(getAppName(true));
		DateFormat format = new SimpleDateFormat(bByMonth != false
												? "yyyyMM"
												: "yyyyMMdd");
		sb.append(format.format(Calendar.getInstance().getTime()));

		// 書きこみストリームを開く
		final FileOutputStream logStream = new FileOutputStream(FileUtilities.combineNameAndExt(sb.toString(), "log"),
																 true);

		Writer writer = null;
		try {
			writer = new BufferedWriter(new OutputStreamWriter(logStream, "UTF8"));
		}
		catch (java.io.UnsupportedEncodingException e) {
			// OutputStreamWriterのコンストラクタででthrows指定されている例外だが、
			// ここで指定している固定文字列では発生しないことが検証済みなので無視する
		}
		return writer;
	}

	//
	// 操作
	//

	// 初期化処理を行う	
	public final boolean initialize(
		ArgumentInfo argument)	// 引数 
	{
		if (argument == null) {
			throw new NullPointerException();
		}

		outputToError("Start.");

		if (properties != null) {
			loadProperties();
		}

		implementor = createImplementor(argument);
		if (implementor != null) {
			if (doInit(argument) != false) {
				implementor.addWindowListener(this);
				restoreState();
			}
			else {
				implementor.dispose();
				implementor = null;
			}
		}

		return (implementor != null);
	}
	public final boolean initialize(
		String[] astrArg,			// 引数
		boolean bIgnoreArgCase)	// 引数大文字小文字区別無視フラグ 
	{
		if (astrArg == null) {
			throw new NullPointerException();
		}

		return initialize(new ArgumentInfo(bIgnoreArgCase, astrArg));
	}
	
	// 表示する
	public final void show()
	{
		if (implementor == null) {
			throw new IllegalStateException();
		}

		implementor.show();
	}

	// メッセージを表示する
	public final void showMessage(
		String strMessage)	// 表示メッセージ
	{
		getOperationBase().showMessage(strMessage);
	}

	//
	// 実装
	//

	// プロパティファイル名を作成する
	private String getPropertiesFileName()
	{
		// ファイル名を作成する
		StringBuffer sb = new StringBuffer();
		sb.append(EnvInfo.getInstance().getUserHomeDir());
		sb.append(File.separatorChar);
		sb.append(getAppName(true));
		sb.append(".properties");
		return sb.toString();
	}

	// プロパティファイルを読みこむ
	private void loadProperties()
	{
		FileInputStream stream = null;
		try {
			File propFile = new File(getPropertiesFileName());
			if (propFile.exists()) {
				stream = new FileInputStream(propFile);
				properties.load(stream);
			}
		}
		catch (IOException exp){
			// Do nothing.
			outputToError("Cannot load propreties file.");
			outputExceptionInfo(exp);
		}
		finally {
			if (stream != null) {
				try {
					stream.close();
				}
				catch (Exception exp) {
					// Do nothing.
					outputExceptionInfo(exp);
				}
			}
		}
	}

	// 状態プロパティを設定する
	private void setStateProperties()
	{
		if (EnvInfo.getInstance().showMainFrameAsFullScreen() == false) { 
			// 現在位置とサイズを保存する
			Point location = implementor.getLocation();
			properties.put(PROPKEY_POS_LEFT, new Integer(location.x).toString());
			properties.put(PROPKEY_POS_TOP, new Integer(location.y).toString());
		
			Dimension size = implementor.getSize();
			properties.put(PROPKEY_POS_WIDTH, new Integer(size.width).toString());
			properties.put(PROPKEY_POS_HEIGHT, new Integer(size.height).toString());
		}

		setAdditionalStatePropreties();
	}

	// プロパティファイルを保存する
	private void storeProperties()
	{
		FileOutputStream stream = null;
		try {
			stream = new FileOutputStream(getPropertiesFileName());
			properties.store(stream, getAppName(false));
		}
		catch (IOException exp){
			// Do nothing.
			outputToError("Cannot store propreties file.");
			outputExceptionInfo(exp);
		}
		finally {
			if (stream != null) {
				try {
					stream.close();
				}
				catch (Exception exp) {
					// Do nothing.
					outputExceptionInfo(exp);
				}
			}
		}
	}

	// 状態を復元する
	private void restoreState()
	{
		boolean bShowAsFullScreen = EnvInfo.getInstance().showMainFrameAsFullScreen(); 
		Dimension screenSize = implementor.getToolkit().getScreenSize();  
		Dimension frameSize = (bShowAsFullScreen
							 ? screenSize
							 : getDefaultSize());
		Point location = new Point((screenSize.width -  frameSize.width) / 2,
									(screenSize.height - frameSize.height) / 2);

		try {
			if (properties != null && bShowAsFullScreen == false) {
				// サイズを復元する
				if (properties.containsKey(PROPKEY_POS_WIDTH)) {
					frameSize.width = Integer.parseInt(properties.getProperty(PROPKEY_POS_WIDTH));
				}
				if (properties.containsKey(PROPKEY_POS_HEIGHT)) {
					frameSize.height = Integer.parseInt(properties.getProperty(PROPKEY_POS_HEIGHT));
				}
	
				// 位置を復元する
				if (properties.containsKey(PROPKEY_POS_LEFT)) {
					location.x = Integer.parseInt(properties.getProperty(PROPKEY_POS_LEFT));
				}
				if (properties.containsKey(PROPKEY_POS_TOP)) {
					location.y = Integer.parseInt(properties.getProperty(PROPKEY_POS_TOP));
				}
			}
		}
		catch (Exception exp) {
			outputToError("Can't restore frame location or size.");
			outputExceptionInfo(exp);
		}

		implementor.setSize(frameSize);
		implementor.setLocation(location);

		restoreAdditionalState();
	}

	// 設定値を返す
	protected final Map getStateProperties()
	{
		if (properties == null) {
			throw new IllegalStateException();
		}

		return Collections.unmodifiableMap(properties);
	}
	
	// 設定値を設定する
	protected final void setStatePropertyValue(
		String strKey,		// キー名
		String strValue)	// 値
	{
		if (properties == null) {
			throw new IllegalStateException();
		}
		if (strKey == null) {
			throw new NullPointerException();
		}
		else if (strKey.length() == 0) {
			throw new IllegalArgumentException();
		}
		if (strValue == null) {
			throw new NullPointerException();
		}

		properties.put(strKey, strValue);
	}

	// 標準エラー出力にメッセージを出力する
	protected final void outputToError(
		String strMsg)	// メッセージ
	{
		Utilities.reportToError(getAppName(false), strMsg);
	}

	// 標準エラー出力にメッセージを出力する
	protected final void outputExceptionInfo(
		Exception exp)	// 例外
	{
		Utilities.reportException(getAppName(false), exp);
	}

	// 実装のインスタンスを返す
	protected final Frame getImplementor()
	{
		return implementor;
	}

	// 実装の基本インターフェースを返す
	private final ImplementOperationBase getOperationBase()
	{
		return (ImplementOperationBase)implementor;
	}

	// システムクリップボードを返す
	protected final Clipboard getSystemClipboard()
	{
		return implementor.getToolkit().getSystemClipboard();
	}

	// フレームにアプリケーションアイコンを設定する
	protected final void setIcon(
		String strIconResourceName)	// アイコンリソース名
	{
		if (EnvInfo.getInstance().isIconRequired() == false) {
			return;
		}

		final URL urlIcon = MainFrameBase.class.getResource(strIconResourceName);
		if (urlIcon != null) {
			implementor.setIconImage(implementor.getToolkit().getImage(urlIcon));
		}
		else {
			outputToError("No icon image file.(at Application#setIcon)");			
		}
	}
	protected final void setIcon()
	{
		setIcon("/icon.png");
	}

	// タイトルを設定する
	protected final void setTitle(
		String strTitle)	// タイトル
	{
		implementor.setTitle(strTitle);
	}

	// カーソルを設定する
	// 注記: 
	protected final void setCursor(
		int nCursorType)	// カーソル種別
	{
		if (saveCursor != null) {
			throw new IllegalStateException();
		}

		saveCursor = implementor.getCursor();
		implementor.setCursor(Cursor.getPredefinedCursor(nCursorType));
	}

	// 待機カーソルを設定する
	protected final void setWaitCursor()
	{
		setCursor(Cursor.WAIT_CURSOR);
	}

	// カーソルを元に戻す
	protected final void restoreCursor()
	{
		if (saveCursor != null) {
			implementor.setCursor(saveCursor);
			saveCursor = null;
		}
	}

	// ビープ音を鳴らす
	protected final void beep()
	{
		implementor.getToolkit().beep();
	}

	// 例外メッセージを表示する
	protected final void showExceptionMessage(
		Exception exp)	// 表示例外
	{
		String strMessage = exp.getMessage();
		if (strMessage == null) {
			strMessage = exp.getMessage();
		}
		if (strMessage == null) {
			strMessage = exp.toString();
		}
		else {
			StringBuffer sb = new StringBuffer();
			sb.append("Exception ocurred : ");
			sb.append(strMessage);
			sb.append("(" + exp.getClass() + ")");
			strMessage = sb.toString();
		}
		getOperationBase().showMessage(strMessage);
		outputExceptionInfo(exp);
	}

	//
	// 再定義メソッド(WindowAdapter)
	//

	// ウィンドウを閉じるための問いあわせ
	public final void windowClosing(
		WindowEvent e)	// イベント情報
	{
		try {
			if (canExit()) {
				implementor.dispose();
			}
		}
		catch (Exception exp) {
			outputExceptionInfo(exp);
		}
	}

	// ウィンドウを閉じる
	public final void windowClosed(
		WindowEvent e)	// イベント情報
	{
		try {
			prepareExit();
		}
		catch (Exception exp) {
			outputExceptionInfo(exp);
		}

		if (properties != null) {
			setStateProperties();

			// 状態をプロパティファイルに保存する
			storeProperties();
		}

		implementor = null;

		outputToError("End.");
		System.exit(0);
	}

	//
	// 再定義可能メソッド(実装)
	//

	// デフォルトのサイズを返す
	protected Dimension getDefaultSize()
	{
		if (implementor == null) {
			throw new IllegalStateException();
		}

		return EnvInfo.getInstance().getDefaultMainFrameSize(implementor);
	}

	// 追加状態を復元する
	protected void restoreAdditionalState()
	{
		// Do nothing on default.
	}

	// 追加状態を設定する
	protected void setAdditionalStatePropreties()
	{
		// Do nothing on default.
	}

	//
	// 再定義必須メソッド(実装)
	//

	// アプリケーション名を返す
	abstract protected String getAppName(boolean bShort);

	// 実装を作成する
	abstract protected Frame createImplementor(ArgumentInfo argument);

	// 初期化処理を行う	
	abstract protected boolean doInit(ArgumentInfo argument);

	// 終了可能かを返す
	abstract protected boolean canExit();

	// 終了準備処理を行う
	abstract protected void prepareExit();
}
